@@ -1,11 +1,10 @@
 from __future__ import absolute_import
 
-import os
+import sys, os, pickle, hashlib, logging
 from io import open
-import pickle
 
 
-from .utils import STRING_TYPE, Serialize, SerializeMemoizer
+from .utils import STRING_TYPE, Serialize, SerializeMemoizer, FS
 from .load_grammar import load_grammar
 from .tree import Tree
 from .common import LexerConf, ParserConf
@@ -22,32 +21,61 @@ class LarkOptions(Serialize):
 
     """
     OPTIONS_DOC = """
-        parser - Decides which parser engine to use, "earley" or "lalr". (Default: "earley")
-                 Note: "lalr" requires a lexer
+# General
 
-        lexer - Decides whether or not to use a lexer stage
-            "standard": Use a standard lexer
-            "contextual": Stronger lexer (only works with parser="lalr")
-            "dynamic": Flexible and powerful (only with parser="earley")
-            "dynamic_complete": Same as dynamic, but tries *every* variation
-                                of tokenizing possible. (only with parser="earley")
-            "auto" (default): Choose for me based on grammar and parser
+    start - The start symbol. Either a string, or a list of strings for
+            multiple possible starts (Default: "start")
+    debug - Display debug information, such as warnings (default: False)
+    transformer - Applies the transformer to every parse tree (equivlent to
+                  applying it after the parse, but faster)
+    propagate_positions - Propagates (line, column, end_line, end_column)
+                          attributes into all tree branches.
+    maybe_placeholders - When True, the `[]` operator returns `None` when not matched.
+                         When `False`,  `[]` behaves like the `?` operator,
+                             and returns no value at all.
+                         (default=`False`. Recommended to set to `True`)
+    cache - Cache the results of the Lark grammar analysis, for x2 to x3 faster loading.
+            LALR only for now.
+        When `False`, does nothing (default)
+        When `True`, caches to a temporary file in the local directory
+        When given a string, caches to the path pointed by the string
 
-        ambiguity - Decides how to handle ambiguity in the parse. Only relevant if parser="earley"
-            "resolve": The parser will automatically choose the simplest derivation
-                       (it chooses consistently: greedy for tokens, non-greedy for rules)
-            "explicit": The parser will return all derivations wrapped in "_ambig" tree nodes (i.e. a forest).
+    g_regex_flags - Flags that are applied to all terminals
+                    (both regex and strings)
+    keep_all_tokens - Prevent the tree builder from automagically
+                      removing "punctuation" tokens (default: False)
 
-        transformer - Applies the transformer to every parse tree
-        debug - Affects verbosity (default: False)
-        keep_all_tokens - Don't automagically remove "punctuation" tokens (default: False)
-        cache_grammar - Cache the Lark grammar (Default: False)
-        postlex - Lexer post-processing (Default: None) Only works with the standard and contextual lexers.
-        start - The start symbol, either a string, or a list of strings for multiple possible starts (Default: "start")
-        priority - How priorities should be evaluated - auto, none, normal, invert (Default: auto)
-        propagate_positions - Propagates [line, column, end_line, end_column] attributes into all tree branches.
-        lexer_callbacks - Dictionary of callbacks for the lexer. May alter tokens during lexing. Use with caution.
-        maybe_placeholders - Experimental feature. Instead of omitting optional rules (i.e. rule?), replace them with None
+# Algorithm
+
+    parser - Decides which parser engine to use
+             Accepts "earley" or "lalr". (Default: "earley")
+             (there is also a "cyk" option for legacy)
+
+    lexer - Decides whether or not to use a lexer stage
+        "auto" (default): Choose for me based on the parser
+        "standard": Use a standard lexer
+        "contextual": Stronger lexer (only works with parser="lalr")
+        "dynamic": Flexible and powerful (only with parser="earley")
+        "dynamic_complete": Same as dynamic, but tries *every* variation
+                            of tokenizing possible.
+
+    ambiguity - Decides how to handle ambiguity in the parse.
+                Only relevant if parser="earley"
+        "resolve": The parser will automatically choose the simplest
+                    derivation (it chooses consistently: greedy for
+                    tokens, non-greedy for rules)
+        "explicit": The parser will return all derivations wrapped
+                    in "_ambig" tree nodes (i.e. a forest).
+
+# Domain Specific
+
+    postlex - Lexer post-processing (Default: None) Only works with the
+                standard and contextual lexers.
+    priority - How priorities should be evaluated - auto, none, normal,
+                invert (Default: auto)
+    lexer_callbacks - Dictionary of callbacks for the lexer. May alter
+                        tokens during lexing. Use with caution.
+    edit_terminals - A callback
     """
     if __doc__:
         __doc__ += OPTIONS_DOC
@@ -56,7 +84,7 @@ class LarkOptions(Serialize):
         'debug': False,
         'keep_all_tokens': False,
         'tree_class': None,
-        'cache_grammar': False,
+        'cache': False,
         'postlex': None,
         'parser': 'earley',
         'lexer': 'auto',
@@ -68,6 +96,7 @@ class LarkOptions(Serialize):
         'lexer_callbacks': {},
         'maybe_placeholders': False,
         'edit_terminals': None,
+        'g_regex_flags': 0,
     }
 
     def __init__(self, options_dict):
@@ -77,7 +106,7 @@ class LarkOptions(Serialize):
         for name, default in self._defaults.items():
             if name in o:
                 value = o.pop(name)
-                if isinstance(default, bool):
+                if isinstance(default, bool) and name != 'cache':
                     value = bool(value)
             else:
                 value = default
@@ -122,6 +151,7 @@ class Lark(Serialize):
             grammar : a string or file-object containing the grammar spec (using Lark's ebnf syntax)
             options : a dictionary controlling various aspects of Lark.
         """
+
         self.options = LarkOptions(options)
 
         # Some, but not all file-like objects have a 'name' attribute
@@ -140,8 +170,27 @@ class Lark(Serialize):
 
         assert isinstance(grammar, STRING_TYPE)
 
-        if self.options.cache_grammar:
-            raise NotImplementedError("Not available yet")
+        cache_fn = None
+        if self.options.cache:
+            if self.options.parser != 'lalr':
+                raise NotImplementedError("cache only works with parser='lalr' for now")
+            if isinstance(self.options.cache, STRING_TYPE):
+                cache_fn = self.options.cache
+            else:
+                if self.options.cache is not True:
+                    raise ValueError("cache must be bool or str")
+                unhashable = ('transformer', 'postlex', 'lexer_callbacks', 'edit_terminals')
+                from . import __version__
+                options_str = ''.join(k+str(v) for k, v in options.items() if k not in unhashable)
+                s = grammar + options_str + __version__
+                md5 = hashlib.md5(s.encode()).hexdigest()
+                cache_fn = '.lark_cache_%s.tmp' % md5
+
+            if FS.exists(cache_fn):
+                logging.debug('Loading grammar from cache: %s', cache_fn)
+                with FS.open(cache_fn, 'rb') as f:
+                    self._load(f, self.options.transformer, self.options.postlex)
+                return
 
         if self.options.lexer == 'auto':
             if self.options.parser == 'lalr':
@@ -209,20 +258,25 @@ class Lark(Serialize):
                 if hasattr(t, term.name):
                     lexer_callbacks[term.name] = getattr(t, term.name)
 
-        self.lexer_conf = LexerConf(self.terminals, self.ignore_tokens, self.options.postlex, lexer_callbacks)
+        self.lexer_conf = LexerConf(self.terminals, self.ignore_tokens, self.options.postlex, lexer_callbacks, self.options.g_regex_flags)
 
         if self.options.parser:
             self.parser = self._build_parser()
         elif lexer:
             self.lexer = self._build_lexer()
 
+        if cache_fn:
+            logging.debug('Saving grammar to cache: %s', cache_fn)
+            with FS.open(cache_fn, 'wb') as f:
+                self.save(f)
+
     if __init__.__doc__:
-        __init__.__doc__ += "\nOPTIONS:" + LarkOptions.OPTIONS_DOC
+        __init__.__doc__ += "\nOptions:\n" + LarkOptions.OPTIONS_DOC
 
     __serialize_fields__ = 'parser', 'rules', 'options'
 
     def _build_lexer(self):
-        return TraditionalLexer(self.lexer_conf.tokens, ignore=self.lexer_conf.ignore, user_callbacks=self.lexer_conf.callbacks)
+        return TraditionalLexer(self.lexer_conf.tokens, ignore=self.lexer_conf.ignore, user_callbacks=self.lexer_conf.callbacks, g_regex_flags=self.lexer_conf.g_regex_flags)
 
     def _prepare_callbacks(self):
         self.parser_class = get_frontend(self.options.parser, self.options.lexer)
@@ -234,34 +288,41 @@ class Lark(Serialize):
         parser_conf = ParserConf(self.rules, self._callbacks, self.options.start)
         return self.parser_class(self.lexer_conf, parser_conf, options=self.options)
 
-    @classmethod
-    def deserialize(cls, data, namespace, memo, transformer=None, postlex=None):
-        if memo:
-            memo = SerializeMemoizer.deserialize(memo, namespace, {})
-        inst = cls.__new__(cls)
-        options = dict(data['options'])
-        if transformer is not None:
-            options['transformer'] = transformer
-        if postlex is not None:
-            options['postlex'] = postlex
-        inst.options = LarkOptions.deserialize(options, memo)
-        inst.rules = [Rule.deserialize(r, memo) for r in data['rules']]
-        inst.source = '<deserialized>'
-        inst._prepare_callbacks()
-        inst.parser = inst.parser_class.deserialize(data['parser'], memo, inst._callbacks, inst.options.postlex)
-        return inst
-
     def save(self, f):
         data, m = self.memo_serialize([TerminalDef, Rule])
         pickle.dump({'data': data, 'memo': m}, f)
 
     @classmethod
     def load(cls, f):
-        d = pickle.load(f)
-        namespace = {'Rule': Rule, 'TerminalDef': TerminalDef}
-        memo = d['memo']
-        return Lark.deserialize(d['data'], namespace, memo)
+        inst = cls.__new__(cls)
+        return inst._load(f)
 
+    def _load(self, f, transformer=None, postlex=None):
+        if isinstance(f, dict):
+            d = f
+        else:
+            d = pickle.load(f)
+        memo = d['memo']
+        data = d['data']
+
+        assert memo
+        memo = SerializeMemoizer.deserialize(memo, {'Rule': Rule, 'TerminalDef': TerminalDef}, {})
+        options = dict(data['options'])
+        if transformer is not None:
+            options['transformer'] = transformer
+        if postlex is not None:
+            options['postlex'] = postlex
+        self.options = LarkOptions.deserialize(options, memo)
+        self.rules = [Rule.deserialize(r, memo) for r in data['rules']]
+        self.source = '<deserialized>'
+        self._prepare_callbacks()
+        self.parser = self.parser_class.deserialize(data['parser'], memo, self._callbacks, self.options.postlex)
+        return self
+
+    @classmethod
+    def _load_from_dict(cls, data, memo, transformer=None, postlex=None):
+        inst = cls.__new__(cls)
+        return inst._load({'data': data, 'memo': memo}, transformer, postlex)
 
     @classmethod
     def open(cls, grammar_filename, rel_to=None, **options):
